######################################################################
# Common constants
######################################################################

NORMAL = 'normal'
LEFT = 'left'
RIGHT = 'right'
TOP = 'top'
BOTTOM = 'bottom'
CENTER = 'center'

COLUMN = 'column'
ROW = 'row'

######################################################################
# Text Justification
######################################################################

JUSTIFY = 'justify'

######################################################################
# Text Direction
######################################################################

RTL = 'rtl'
LTR = 'ltr'

######################################################################
# Font family
######################################################################

SERIF = 'serif'
SANS_SERIF = 'sans-serif'
CURSIVE = 'cursive'
FANTASY = 'fantasy'
MONOSPACE = 'monospace'

######################################################################
# Font Styling
######################################################################

ITALIC = 'italic'
OBLIQUE = 'oblique'

######################################################################
# Font Variant
######################################################################

SMALL_CAPS = 'small-caps'

######################################################################
# Font boldness
######################################################################

BOLD = 'bold'

######################################################################
# Colors
######################################################################

TRANSPARENT = 'transparent'

ALICEBLUE = 'aliceblue'
ANTIQUEWHITE = 'antiquewhite'
AQUA = 'aqua'
AQUAMARINE = 'aquamarine'
AZURE = 'azure'
BEIGE = 'beige'
BISQUE = 'bisque'
BLACK = 'black'
BLANCHEDALMOND = 'blanchedalmond'
BLUE = 'blue'
BLUEVIOLET = 'blueviolet'
BROWN = 'brown'
BURLYWOOD = 'burlywood'
CADETBLUE = 'cadetblue'
CHARTREUSE = 'chartreuse'
CHOCOLATE = 'chocolate'
CORAL = 'coral'
CORNFLOWERBLUE = 'cornflowerblue'
CORNSILK = 'cornsilk'
CRIMSON = 'crimson'
CYAN = 'cyan'
DARKBLUE = 'darkblue'
DARKCYAN = 'darkcyan'
DARKGOLDENROD = 'darkgoldenrod'
DARKGRAY = 'darkgray'
DARKGREY = 'darkgrey'
DARKGREEN = 'darkgreen'
DARKKHAKI = 'darkkhaki'
DARKMAGENTA = 'darkmagenta'
DARKOLIVEGREEN = 'darkolivegreen'
DARKORANGE = 'darkorange'
DARKORCHID = 'darkorchid'
DARKRED = 'darkred'
DARKSALMON = 'darksalmon'
DARKSEAGREEN = 'darkseagreen'
DARKSLATEBLUE = 'darkslateblue'
DARKSLATEGRAY = 'darkslategray'
DARKSLATEGREY = 'darkslategrey'
DARKTURQUOISE = 'darkturquoise'
DARKVIOLET = 'darkviolet'
DEEPPINK = 'deeppink'
DEEPSKYBLUE = 'deepskyblue'
DIMGRAY = 'dimgray'
DIMGREY = 'dimgrey'
DODGERBLUE = 'dodgerblue'
FIREBRICK = 'firebrick'
FLORALWHITE = 'floralwhite'
FORESTGREEN = 'forestgreen'
FUCHSIA = 'fuchsia'
GAINSBORO = 'gainsboro'
GHOSTWHITE = 'ghostwhite'
GOLD = 'gold'
GOLDENROD = 'goldenrod'
GRAY = 'gray'
GREY = 'grey'
GREEN = 'green'
GREENYELLOW = 'greenyellow'
HONEYDEW = 'honeydew'
HOTPINK = 'hotpink'
INDIANRED = 'indianred'
INDIGO = 'indigo'
IVORY = 'ivory'
KHAKI = 'khaki'
LAVENDER = 'lavender'
LAVENDERBLUSH = 'lavenderblush'
LAWNGREEN = 'lawngreen'
LEMONCHIFFON = 'lemonchiffon'
LIGHTBLUE = 'lightblue'
LIGHTCORAL = 'lightcoral'
LIGHTCYAN = 'lightcyan'
LIGHTGOLDENRODYELLOW = 'lightgoldenrodyellow'
LIGHTGRAY = 'lightgray'
LIGHTGREY = 'lightgrey'
LIGHTGREEN = 'lightgreen'
LIGHTPINK = 'lightpink'
LIGHTSALMON = 'lightsalmon'
LIGHTSEAGREEN = 'lightseagreen'
LIGHTSKYBLUE = 'lightskyblue'
LIGHTSLATEGRAY = 'lightslategray'
LIGHTSLATEGREY = 'lightslategrey'
LIGHTSTEELBLUE = 'lightsteelblue'
LIGHTYELLOW = 'lightyellow'
LIME = 'lime'
LIMEGREEN = 'limegreen'
LINEN = 'linen'
MAGENTA = 'magenta'
MAROON = 'maroon'
MEDIUMAQUAMARINE = 'mediumaquamarine'
MEDIUMBLUE = 'mediumblue'
MEDIUMORCHID = 'mediumorchid'
MEDIUMPURPLE = 'mediumpurple'
MEDIUMSEAGREEN = 'mediumseagreen'
MEDIUMSLATEBLUE = 'mediumslateblue'
MEDIUMSPRINGGREEN = 'mediumspringgreen'
MEDIUMTURQUOISE = 'mediumturquoise'
MEDIUMVIOLETRED = 'mediumvioletred'
MIDNIGHTBLUE = 'midnightblue'
MINTCREAM = 'mintcream'
MISTYROSE = 'mistyrose'
MOCCASIN = 'moccasin'
NAVAJOWHITE = 'navajowhite'
NAVY = 'navy'
OLDLACE = 'oldlace'
OLIVE = 'olive'
OLIVEDRAB = 'olivedrab'
ORANGE = 'orange'
ORANGERED = 'orangered'
ORCHID = 'orchid'
PALEGOLDENROD = 'palegoldenrod'
PALEGREEN = 'palegreen'
PALETURQUOISE = 'paleturquoise'
PALEVIOLETRED = 'palevioletred'
PAPAYAWHIP = 'papayawhip'
PEACHPUFF = 'peachpuff'
PERU = 'peru'
PINK = 'pink'
PLUM = 'plum'
POWDERBLUE = 'powderblue'
PURPLE = 'purple'
REBECCAPURPLE = 'rebeccapurple'
RED = 'red'
ROSYBROWN = 'rosybrown'
ROYALBLUE = 'royalblue'
SADDLEBROWN = 'saddlebrown'
SALMON = 'salmon'
SANDYBROWN = 'sandybrown'
SEAGREEN = 'seagreen'
SEASHELL = 'seashell'
SIENNA = 'sienna'
SILVER = 'silver'
SKYBLUE = 'skyblue'
SLATEBLUE = 'slateblue'
SLATEGRAY = 'slategray'
SLATEGREY = 'slategrey'
SNOW = 'snow'
SPRINGGREEN = 'springgreen'
STEELBLUE = 'steelblue'
TAN = 'tan'
TEAL = 'teal'
THISTLE = 'thistle'
TOMATO = 'tomato'
TURQUOISE = 'turquoise'
VIOLET = 'violet'
WHEAT = 'wheat'
WHITE = 'white'
WHITESMOKE = 'whitesmoke'
YELLOW = 'yellow'
YELLOWGREEN = 'yellowgreen'




class at_least:
    "An annotation to wrap around a value to describe that it is a minimum bound"
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return 'at least {0}'.format(self.value)

    def __eq__(self, other):
        try:
            return self.value == other.value
        except AttributeError:
            return False


class BaseIntrinsicSize:
    """Representation of the intrinsic size of an object.

    width: The width of the node.
    height: The height of the node.
    ratio: The height between height and width. width = height * ratio
    """
    def __init__(self, width=None, height=None, ratio=None, layout=None):
        self._layout = layout
        self._width = width
        self._height = height

        self._ratio = None

    def __repr__(self):
        return '({}, {})'.format(self.width, self.height)

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        if self._width != value:
            self._width = value

            if self._layout:
                self._layout.dirty(intrinsic_width=value)

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value):
        if self._height != value:
            self._height = value

            if self._layout:
                self._layout.dirty(intrinsic_height=value)

    @property
    def ratio(self):
        return self._ratio

    @ratio.setter
    def ratio(self, value):
        if self._ratio != value:
            self._ratio = value
            if self._layout:
                self._layout.dirty(intrinsic_ratio=value)





def set_root(node, root):
    # Propegate a root node change through a tree.
    node._root = root
    for child in node.children:
        set_root(child, root)


class Node:
    def __init__(self, style, applicator=None, children=None):
        self.applicator = applicator
        self.style = style.copy(applicator)
        self.intrinsic = self.style.IntrinsicSize()
        self.layout = self.style.Box(self)

        self._parent = None
        self._root = None
        if children is None:
            self._children = None
        else:
            self._children = []
            for child in children:
                self.add(child)

    @property
    def root(self):
        """The root of the tree containing this node.

        Returns:
            The root node. Returns self if this node *is* the root node.
        """
        return self._root if self._root else self

    @property
    def parent(self):
        """ The parent of this node.

        Returns:
            The parent of this node. Returns None if this node is the root node.
        """
        return self._parent

    @property
    def children(self):
        """ The children of this node.
        This *always* returns a list, even if the node is a leaf
        and cannot have children.

        Returns:
            A list of the children for this widget.
        """
        if self._children is None:
            return []
        else:
            return self._children

    @property
    def can_have_children(self):
        """Determine if the node can have children.

        This does not resolve whether there actually *are* any children;
        it only confirms whether children are theoretically allowed.
        """
        return self._children is not None

    def add(self, child):
        """Add a node as a child of this one.
        Args:
            child: A node to add as a child to this node.

        Raises:
            ValueError: If this node is a leaf, and cannot have children.
        """
        if self._children is None:
            raise ValueError('Cannot add children')

        self._children.append(child)
        child._parent = self
        set_root(child, self.root)

    def refresh(self, viewport):
        """Refresh the layout and appearance of the tree this node is contained in."""
        if self._root:
            self._root.refresh(viewport)
        else:
            self.style.layout(self, viewport)
            if self.applicator:
                self.applicator.set_bounds()



from unittest import TestCase

from travertino.declaration import BaseStyle
from travertino.layout import BaseBox, Viewport
from travertino.node import Node
from travertino.size import BaseIntrinsicSize


class Style(BaseStyle):
    class IntrinsicSize(BaseIntrinsicSize):
        pass

    class Box(BaseBox):
        pass

    def layout(self, root, viewport):
        # A simple layout scheme that allocats twice the viewport size.
        root.layout.content_width = viewport.width * 2
        root.layout.content_height = viewport.height * 2


class ViewportTests(TestCase):
    def test_default(self):
        viewport = Viewport()

        self.assertEqual(viewport.width, 0)
        self.assertEqual(viewport.height, 0)
        self.assertEqual(viewport.dpi, None)

    def test_constructor(self):
        viewport = Viewport(width=640, height=480, dpi=96)

        self.assertEqual(viewport.width, 640)
        self.assertEqual(viewport.height, 480)
        self.assertEqual(viewport.dpi, 96)


class BoxTests(TestCase):
    def setUp(self):
        self.maxDiff = None

        self.grandchild1_1 = Node(style=Style())
        self.grandchild1_1.layout.content_width = 10
        self.grandchild1_1.layout.content_height = 16
        self.grandchild1_2 = Node(style=Style())

        self.child1 = Node(style=Style(), children=[self.grandchild1_1, self.grandchild1_2])
        self.child1.layout.content_width = 10
        self.child1.layout.content_height = 16
        self.child2 = Node(style=Style(), children=[])

        self.node = Node(style=Style(), children=[self.child1, self.child2])
        self.node.layout.content_width = 10
        self.node.layout.content_height = 16

    def assertLayout(self, box, expected):
        actual = {
            'origin': (box._origin_left, box._origin_top),
            'size': (box.width, box.height),
            'content': (box.content_width, box.content_height),
            'relative': (
                box.content_top,
                box.content_right,
                box.content_bottom,
                box.content_left,
            ),
            'absolute': (
                box.absolute_content_top,
                box.absolute_content_right,
                box.absolute_content_bottom,
                box.absolute_content_left,
            ),
        }
        self.assertEqual(actual, expected)

    def test_repr(self):
        self.node.layout._origin_top = 1
        self.node.layout._origin_left = 2
        self.assertEqual(repr(self.node.layout), "<Box (10x16 @ 2,1)>")

    def test_initial(self):
        # Core attributes have been stored
        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (10, 16),
                'content': (10, 16),
                'relative': (0, 0, 0, 0),
                'absolute': (0, 10, 16, 0),
            }
        )

    def test_set_content_top(self):
        self.node.layout.content_top = 5

        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (10, 21),
                'content': (10, 16),
                'relative': (5, 0, 0, 0),
                'absolute': (5, 10, 21, 0),
            }
        )

        # Set the top to a new value
        self.node.layout.content_top = 7

        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (10, 23),
                'content': (10, 16),
                'relative': (7, 0, 0, 0),
                'absolute': (7, 10, 23, 0),
            }
        )

    def test_set_content_left(self):
        self.node.layout.content_left = 5

        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (15, 16),
                'content': (10, 16),
                'relative': (0, 0, 0, 5),
                'absolute': (0, 15, 16, 5),
            }
        )

        # Set the left to a new value
        self.node.layout.content_left = 7

        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (17, 16),
                'content': (10, 16),
                'relative': (0, 0, 0, 7),
                'absolute': (0, 17, 16, 7),
            }
        )

    def test_set_content_width(self):
        self.node.layout.content_width = 5

        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (5, 16),
                'content': (5, 16),
                'relative': (0, 0, 0, 0),
                'absolute': (0, 5, 16, 0),
            }
        )

        # Set the width to a new value
        self.node.layout.content_width = 7

        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (7, 16),
                'content': (7, 16),
                'relative': (0, 0, 0, 0),
                'absolute': (0, 7, 16, 0),
            }
        )

    def test_set_content_height(self):
        self.node.layout.content_height = 5

        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (10, 5),
                'content': (10, 5),
                'relative': (0, 0, 0, 0),
                'absolute': (0, 10, 5, 0),
            }
        )

        # Set the height to a new value
        self.node.layout.content_height = 7

        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (10, 7),
                'content': (10, 7),
                'relative': (0, 0, 0, 0),
                'absolute': (0, 10, 7, 0),
            }
        )

    def test_descendent_offsets(self):
        self.node.layout.content_top = 7
        self.node.layout.content_left = 8

        self.child1.layout.content_top = 9
        self.child1.layout.content_left = 10

        self.grandchild1_1.layout.content_top = 11
        self.grandchild1_1.layout.content_left = 12

        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (18, 23),
                'content': (10, 16),
                'relative': (7, 0, 0, 8),
                'absolute': (7, 18, 23, 8),
            }
        )

        self.assertLayout(
            self.child1.layout,
            {
                'origin': (8, 7),
                'size': (20, 25),
                'content': (10, 16),
                'relative': (9, 0, 0, 10),
                'absolute': (16, 28, 32, 18),
            }
        )

        self.assertLayout(
            self.grandchild1_1.layout,
            {
                'origin': (18, 16),
                'size': (22, 27),
                'content': (10, 16),
                'relative': (11, 0, 0, 12),
                'absolute': (27, 40, 43, 30),
            }
        )

        # Modify the grandchild position
        self.grandchild1_1.layout.content_top = 13
        self.grandchild1_1.layout.content_left = 14

        # Only the grandchild position has changed.
        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (18, 23),
                'content': (10, 16),
                'relative': (7, 0, 0, 8),
                'absolute': (7, 18, 23, 8),
            }
        )

        self.assertLayout(
            self.child1.layout,
            {
                'origin': (8, 7),
                'size': (20, 25),
                'content': (10, 16),
                'relative': (9, 0, 0, 10),
                'absolute': (16, 28, 32, 18),
            }
        )

        self.assertLayout(
            self.grandchild1_1.layout,
            {
                'origin': (18, 16),
                'size': (24, 29),
                'content': (10, 16),
                'relative': (13, 0, 0, 14),
                'absolute': (29, 42, 45, 32),
            }
        )

        # Modify the child position
        self.child1.layout.content_top = 15
        self.child1.layout.content_left = 16

        # The child and grandchild position has changed.
        self.assertLayout(
            self.node.layout,
            {
                'origin': (0, 0),
                'size': (18, 23),
                'content': (10, 16),
                'relative': (7, 0, 0, 8),
                'absolute': (7, 18, 23, 8),
            }
        )

        self.assertLayout(
            self.child1.layout,
            {
                'origin': (8, 7),
                'size': (26, 31),
                'content': (10, 16),
                'relative': (15, 0, 0, 16),
                'absolute': (22, 34, 38, 24),
            }
        )

        self.assertLayout(
            self.grandchild1_1.layout,
            {
                'origin': (24, 22),
                'size': (24, 29),
                'content': (10, 16),
                'relative': (13, 0, 0, 14),
                'absolute': (35, 48, 51, 38),
            }
        )

    def test_absolute_equalities(self):
        # Move the box around and set some borders.
        self.node.layout.origin_top = 100
        self.node.layout.origin_left = 200

        self.node.layout.content_top = 50
        self.node.layout.content_left = 75
        self.node.layout.content_right = 42
        self.node.layout.content_bottom = 37

        self.assertEqual(
            self.node.layout.absolute_content_left + self.node.layout.content_width,
            self.node.layout.absolute_content_right
        )
        self.assertEqual(
            self.node.layout.absolute_content_top + self.node.layout.content_height,
            self.node.layout.absolute_content_bottom
        )

        self.assertEqual(
            self.node.layout.content_left + self.node.layout.content_width + self.node.layout.content_right,
            self.node.layout.width
        )
        self.assertEqual(
            self.node.layout.content_top + self.node.layout.content_height + self.node.layout.content_bottom,
            self.node.layout.height
        )



from .constants import *  # NOQA

class Color:
    "A base class for all colorspace representations"
    pass

    def __eq__(self, other):
        try:
            c1 = self.rgba
            c2 = other.rgba

            return (
                c1.r == c2.r
                and c1.g == c2.g
                and c1.b == c2.b
                and c1.a == c2.a
            )
        except AttributeError:
            return False


class rgba(Color):
    "A representation of an RGBA color"
    def __init__(self, r, g, b, a):
        self.r = r
        self.g = g
        self.b = b
        self.a = a

    def __repr__(self):
        return "rgba({}, {}, {}, {})".format(self.r, self.g, self.b, self.a)

    @property
    def rgba(self):
        return self


class rgb(rgba):
    "A representation of an RGB color"
    def __init__(self, r, g, b):
        super().__init__(r, g, b, 1.0)

    def __repr__(self):
        return "rgb({}, {}, {})".format(self.r, self.g, self.b)


class hsla(Color):
    "A representation of an HSLA color"
    def __init__(self, h, s, l, a=1.0):
        self.h = h
        self.s = s
        self.l = l
        self.a = a

    def __repr__(self):
        return "hsla({}, {}, {}, {})".format(self.h, self.s, self.l, self.a)

    @property
    def rgba(self):
        c = (1.0 - abs(2.0 * self.l - 1.0)) * self.s
        h = self.h / 60.0
        x = c * (1.0 - abs(h % 2 - 1.0))
        m = self.l - 0.5 * c

        if h < 1.0:
            r, g, b = c + m, x + m, m
        elif h < 2.0:
            r, g, b = x + m, c + m, m
        elif h < 3.0:
            r, g, b = m, c + m, x + m
        elif h < 4.0:
            r, g, b = m, x + m, c + m
        elif h < 5.0:
            r, g, b = m, x + m, c + m
        else:
            r, g, b = c + m, m, x + m

        return rgba(
            round(r * 0xff),
            round(g * 0xff),
            round(b * 0xff),
            self.a
        )


class hsl(hsla):
    "A representation of an HSL color"
    def __init__(self, h, s, l):
        super().__init__(h, s, l, 1.0)

    def __repr__(self):
        return "hsl({}, {}, {})".format(self.h, self.s, self.l)


def color(value):
    """Parse a color from a value.

    Accepts:
    * rgb() instances
    * hsl() instances
    * '#rgb'
    * '#rgba'
    * '#rrggbb'
    * '#rrggbbaa'
    * '#RGB'
    * '#RGBA'
    * '#RRGGBB'
    * '#RRGGBBAA'
    * 'rgb(0, 0, 0)'
    * 'rgba(0, 0, 0, 0.0)'
    * 'hsl(0, 0%, 0%)'
    * 'hsla(0, 0%, 0%, 0.0)'
    * A named color
    """

    if isinstance(value, (rgb, rgba, hsl, hsla)):
        return value

    elif isinstance(value, str):
        if value[0] == '#':
            if len(value) == 4:
                return rgb(
                    r=int(value[1] + value[1], 16),
                    g=int(value[2] + value[2], 16),
                    b=int(value[3] + value[3], 16),
                )
            elif len(value) == 5:
                return rgba(
                    r=int(value[1] + value[1], 16),
                    g=int(value[2] + value[2], 16),
                    b=int(value[3] + value[3], 16),
                    a=int(value[4] + value[4], 16) / 0xff,
                )
            elif len(value) == 7:
                return rgb(
                    r=int(value[1:3], 16),
                    g=int(value[3:5], 16),
                    b=int(value[5:7], 16),
                )
            elif len(value) == 9:
                return rgba(
                    r=int(value[1:3], 16),
                    g=int(value[3:5], 16),
                    b=int(value[5:7], 16),
                    a=int(value[7:9], 16) / 0xff,
                )
        elif value.startswith('rgba'):
            try:
                values = value[5:-1].split(',')
                if len(values) == 4:
                    return rgba(int(values[0]), int(values[1]), int(values[2]), float(values[3]))
            except ValueError:
                pass
        elif value.startswith('rgb'):
            try:
                values = value[4:-1].split(',')
                if len(values) == 3:
                    return rgb(int(values[0]), int(values[1]), int(values[2]))
            except ValueError:
                pass

        elif value.startswith('hsla'):
            try:
                values = value[5:-1].split(',')
                if len(values) == 4:
                    return hsla(
                        int(values[0]),
                        int(values[1].strip().rstrip('%')) / 100.0,
                        int(values[2].strip().rstrip('%')) / 100.0,
                        float(values[3])
                    )
            except ValueError:
                pass

        elif value.startswith('hsl'):
            try:
                values = value[4:-1].split(',')
                if len(values) == 3:
                    return hsl(
                        int(values[0]),
                        int(values[1].strip().rstrip('%')) / 100.0,
                        int(values[2].strip().rstrip('%')) / 100.0,
                    )
            except ValueError:
                pass
        else:
            try:
                return NAMED_COLOR[value.lower()]
            except KeyError:
                pass

    raise ValueError('Unknown color %s' % value)


NAMED_COLOR = {
    ALICEBLUE: rgb(0xF0, 0xF8, 0xFF),
    ANTIQUEWHITE: rgb(0xFA, 0xEB, 0xD7),
    AQUA: rgb(0x00, 0xFF, 0xFF),
    AQUAMARINE: rgb(0x7F, 0xFF, 0xD4),
    AZURE: rgb(0xF0, 0xFF, 0xFF),
    BEIGE: rgb(0xF5, 0xF5, 0xDC),
    BISQUE: rgb(0xFF, 0xE4, 0xC4),
    BLACK: rgb(0x00, 0x00, 0x00),
    BLANCHEDALMOND: rgb(0xFF, 0xEB, 0xCD),
    BLUE: rgb(0x00, 0x00, 0xFF),
    BLUEVIOLET: rgb(0x8A, 0x2B, 0xE2),
    BROWN: rgb(0xA5, 0x2A, 0x2A),
    BURLYWOOD: rgb(0xDE, 0xB8, 0x87),
    CADETBLUE: rgb(0x5F, 0x9E, 0xA0),
    CHARTREUSE: rgb(0x7F, 0xFF, 0x00),
    CHOCOLATE: rgb(0xD2, 0x69, 0x1E),
    CORAL: rgb(0xFF, 0x7F, 0x50),
    CORNFLOWERBLUE: rgb(0x64, 0x95, 0xED),
    CORNSILK: rgb(0xFF, 0xF8, 0xDC),
    CRIMSON: rgb(0xDC, 0x14, 0x3C),
    CYAN: rgb(0x00, 0xFF, 0xFF),
    DARKBLUE: rgb(0x00, 0x00, 0x8B),
    DARKCYAN: rgb(0x00, 0x8B, 0x8B),
    DARKGOLDENROD: rgb(0xB8, 0x86, 0x0B),
    DARKGRAY: rgb(0xA9, 0xA9, 0xA9),
    DARKGREY: rgb(0xA9, 0xA9, 0xA9),
    DARKGREEN: rgb(0x00, 0x64, 0x00),
    DARKKHAKI: rgb(0xBD, 0xB7, 0x6B),
    DARKMAGENTA: rgb(0x8B, 0x00, 0x8B),
    DARKOLIVEGREEN: rgb(0x55, 0x6B, 0x2F),
    DARKORANGE: rgb(0xFF, 0x8C, 0x00),
    DARKORCHID: rgb(0x99, 0x32, 0xCC),
    DARKRED: rgb(0x8B, 0x00, 0x00),
    DARKSALMON: rgb(0xE9, 0x96, 0x7A),
    DARKSEAGREEN: rgb(0x8F, 0xBC, 0x8F),
    DARKSLATEBLUE: rgb(0x48, 0x3D, 0x8B),
    DARKSLATEGRAY: rgb(0x2F, 0x4F, 0x4F),
    DARKSLATEGREY: rgb(0x2F, 0x4F, 0x4F),
    DARKTURQUOISE: rgb(0x00, 0xCE, 0xD1),
    DARKVIOLET: rgb(0x94, 0x00, 0xD3),
    DEEPPINK: rgb(0xFF, 0x14, 0x93),
    DEEPSKYBLUE: rgb(0x00, 0xBF, 0xFF),
    DIMGRAY: rgb(0x69, 0x69, 0x69),
    DIMGREY: rgb(0x69, 0x69, 0x69),
    DODGERBLUE: rgb(0x1E, 0x90, 0xFF),
    FIREBRICK: rgb(0xB2, 0x22, 0x22),
    FLORALWHITE: rgb(0xFF, 0xFA, 0xF0),
    FORESTGREEN: rgb(0x22, 0x8B, 0x22),
    FUCHSIA: rgb(0xFF, 0x00, 0xFF),
    GAINSBORO: rgb(0xDC, 0xDC, 0xDC),
    GHOSTWHITE: rgb(0xF8, 0xF8, 0xFF),
    GOLD: rgb(0xFF, 0xD7, 0x00),
    GOLDENROD: rgb(0xDA, 0xA5, 0x20),
    GRAY: rgb(0x80, 0x80, 0x80),
    GREY: rgb(0x80, 0x80, 0x80),
    GREEN: rgb(0x00, 0x80, 0x00),
    GREENYELLOW: rgb(0xAD, 0xFF, 0x2F),
    HONEYDEW: rgb(0xF0, 0xFF, 0xF0),
    HOTPINK: rgb(0xFF, 0x69, 0xB4),
    INDIANRED: rgb(0xCD, 0x5C, 0x5C),
    INDIGO: rgb(0x4B, 0x00, 0x82),
    IVORY: rgb(0xFF, 0xFF, 0xF0),
    KHAKI: rgb(0xF0, 0xE6, 0x8C),
    LAVENDER: rgb(0xE6, 0xE6, 0xFA),
    LAVENDERBLUSH: rgb(0xFF, 0xF0, 0xF5),
    LAWNGREEN: rgb(0x7C, 0xFC, 0x00),
    LEMONCHIFFON: rgb(0xFF, 0xFA, 0xCD),
    LIGHTBLUE: rgb(0xAD, 0xD8, 0xE6),
    LIGHTCORAL: rgb(0xF0, 0x80, 0x80),
    LIGHTCYAN: rgb(0xE0, 0xFF, 0xFF),
    LIGHTGOLDENRODYELLOW: rgb(0xFA, 0xFA, 0xD2),
    LIGHTGRAY: rgb(0xD3, 0xD3, 0xD3),
    LIGHTGREY: rgb(0xD3, 0xD3, 0xD3),
    LIGHTGREEN: rgb(0x90, 0xEE, 0x90),
    LIGHTPINK: rgb(0xFF, 0xB6, 0xC1),
    LIGHTSALMON: rgb(0xFF, 0xA0, 0x7A),
    LIGHTSEAGREEN: rgb(0x20, 0xB2, 0xAA),
    LIGHTSKYBLUE: rgb(0x87, 0xCE, 0xFA),
    LIGHTSLATEGRAY: rgb(0x77, 0x88, 0x99),
    LIGHTSLATEGREY: rgb(0x77, 0x88, 0x99),
    LIGHTSTEELBLUE: rgb(0xB0, 0xC4, 0xDE),
    LIGHTYELLOW: rgb(0xFF, 0xFF, 0xE0),
    LIME: rgb(0x00, 0xFF, 0x00),
    LIMEGREEN: rgb(0x32, 0xCD, 0x32),
    LINEN: rgb(0xFA, 0xF0, 0xE6),
    MAGENTA: rgb(0xFF, 0x00, 0xFF),
    MAROON: rgb(0x80, 0x00, 0x00),
    MEDIUMAQUAMARINE: rgb(0x66, 0xCD, 0xAA),
    MEDIUMBLUE: rgb(0x00, 0x00, 0xCD),
    MEDIUMORCHID: rgb(0xBA, 0x55, 0xD3),
    MEDIUMPURPLE: rgb(0x93, 0x70, 0xDB),
    MEDIUMSEAGREEN: rgb(0x3C, 0xB3, 0x71),
    MEDIUMSLATEBLUE: rgb(0x7B, 0x68, 0xEE),
    MEDIUMSPRINGGREEN: rgb(0x00, 0xFA, 0x9A),
    MEDIUMTURQUOISE: rgb(0x48, 0xD1, 0xCC),
    MEDIUMVIOLETRED: rgb(0xC7, 0x15, 0x85),
    MIDNIGHTBLUE: rgb(0x19, 0x19, 0x70),
    MINTCREAM: rgb(0xF5, 0xFF, 0xFA),
    MISTYROSE: rgb(0xFF, 0xE4, 0xE1),
    MOCCASIN: rgb(0xFF, 0xE4, 0xB5),
    NAVAJOWHITE: rgb(0xFF, 0xDE, 0xAD),
    NAVY: rgb(0x00, 0x00, 0x80),
    OLDLACE: rgb(0xFD, 0xF5, 0xE6),
    OLIVE: rgb(0x80, 0x80, 0x00),
    OLIVEDRAB: rgb(0x6B, 0x8E, 0x23),
    ORANGE: rgb(0xFF, 0xA5, 0x00),
    ORANGERED: rgb(0xFF, 0x45, 0x00),
    ORCHID: rgb(0xDA, 0x70, 0xD6),
    PALEGOLDENROD: rgb(0xEE, 0xE8, 0xAA),
    PALEGREEN: rgb(0x98, 0xFB, 0x98),
    PALETURQUOISE: rgb(0xAF, 0xEE, 0xEE),
    PALEVIOLETRED: rgb(0xDB, 0x70, 0x93),
    PAPAYAWHIP: rgb(0xFF, 0xEF, 0xD5),
    PEACHPUFF: rgb(0xFF, 0xDA, 0xB9),
    PERU: rgb(0xCD, 0x85, 0x3F),
    PINK: rgb(0xFF, 0xC0, 0xCB),
    PLUM: rgb(0xDD, 0xA0, 0xDD),
    POWDERBLUE: rgb(0xB0, 0xE0, 0xE6),
    PURPLE: rgb(0x80, 0x00, 0x80),
    REBECCAPURPLE: rgb(0x66, 0x33, 0x99),
    RED: rgb(0xFF, 0x00, 0x00),
    ROSYBROWN: rgb(0xBC, 0x8F, 0x8F),
    ROYALBLUE: rgb(0x41, 0x69, 0xE1),
    SADDLEBROWN: rgb(0x8B, 0x45, 0x13),
    SALMON: rgb(0xFA, 0x80, 0x72),
    SANDYBROWN: rgb(0xF4, 0xA4, 0x60),
    SEAGREEN: rgb(0x2E, 0x8B, 0x57),
    SEASHELL: rgb(0xFF, 0xF5, 0xEE),
    SIENNA: rgb(0xA0, 0x52, 0x2D),
    SILVER: rgb(0xC0, 0xC0, 0xC0),
    SKYBLUE: rgb(0x87, 0xCE, 0xEB),
    SLATEBLUE: rgb(0x6A, 0x5A, 0xCD),
    SLATEGRAY: rgb(0x70, 0x80, 0x90),
    SLATEGREY: rgb(0x70, 0x80, 0x90),
    SNOW: rgb(0xFF, 0xFA, 0xFA),
    SPRINGGREEN: rgb(0x00, 0xFF, 0x7F),
    STEELBLUE: rgb(0x46, 0x82, 0xB4),
    TAN: rgb(0xD2, 0xB4, 0x8C),
    TEAL: rgb(0x00, 0x80, 0x80),
    THISTLE: rgb(0xD8, 0xBF, 0xD8),
    TOMATO: rgb(0xFF, 0x63, 0x47),
    TURQUOISE: rgb(0x40, 0xE0, 0xD0),
    VIOLET: rgb(0xEE, 0x82, 0xEE),
    WHEAT: rgb(0xF5, 0xDE, 0xB3),
    WHITE: rgb(0xFF, 0xFF, 0xFF),
    WHITESMOKE: rgb(0xF5, 0xF5, 0xF5),
    YELLOW: rgb(0xFF, 0xFF, 0x00),
    YELLOWGREEN: rgb(0x9A, 0xCD, 0x32),
}



from .colors import color


class Choices:
    "A class to define allowable data types for a property"
    def __init__(
            self, *constants, default=False,
            string=False, integer=False, number=False, color=False):
        self.constants = set(constants)
        self.default = default

        self.string = string
        self.integer = integer
        self.number = number
        self.color = color

        self._options = sorted(str(c).lower().replace('_', '-') for c in self.constants)
        if self.string:
            self._options.append("<string>")
        if self.integer:
            self._options.append("<integer>")
        if self.number:
            self._options.append("<number>")
        if self.color:
            self._options.append("<color>")

    def validate(self, value):
        if self.default:
            if value is None:
                return None
        if self.string:
            try:
                return value.strip()
            except AttributeError:
                pass
        if self.integer:
            try:
                return int(value)
            except (ValueError, TypeError):
                pass
        if self.number:
            try:
                return float(value)
            except (ValueError, TypeError):
                pass
        if self.color:
            try:
                return color(value)
            except ValueError:
                pass
        if value == 'none':
            value = None
        for const in self.constants:
            if value == const:
                return const

        raise ValueError("'{0}' is not a valid initial value".format(value))

    def __str__(self):
        return ", ".join(self._options)


class BaseStyle:
    """A base class for style declarations.

    Exposes a dict-like interface.
    """
    _PROPERTIES = {}
    _ALL_PROPERTIES = {}

    def __init__(self, **style):
        self._applicator = None
        self.update(**style)

    ######################################################################
    # Interface that style declarations must define
    ######################################################################

    def apply(self, property, value):
        raise NotImplementedError('Style must define an apply method')  # pragma: no cover

    ######################################################################
    # Provide a dict-like interface
    ######################################################################

    def reapply(self):
        for style in self._PROPERTIES.get(self.__class__, set()):
            self.apply(style, getattr(self, style))

    def update(self, **styles):
        "Set multiple styles on the style definition."
        for name, value in styles.items():
            name = name.replace('-', '_')
            if not name in self._ALL_PROPERTIES.get(self.__class__, set()):
                raise NameError("Unknown style '%s'" % name)

            setattr(self, name, value)

    def copy(self, applicator=None):
        "Create a duplicate of this style declaration."
        dup = self.__class__()
        dup._applicator = applicator
        for style in self._PROPERTIES.get(self.__class__, set()):
            try:
                setattr(dup, style, getattr(self, '_%s' % style))
            except AttributeError:
                pass
        return dup

    def __getitem__(self, name):
        name = name.replace('-', '_')
        if name in self._PROPERTIES.get(self.__class__, set()):
            return getattr(self, name)
        raise KeyError(name)

    def __setitem__(self, name, value):
        name = name.replace('-', '_')
        if name in self._PROPERTIES.get(self.__class__, set()):
            setattr(self, name, value)
        else:
            raise KeyError(name)

    def __delitem__(self, name):
        name = name.replace('-', '_')
        if name in self._PROPERTIES.get(self.__class__, set()):
            delattr(self, name)
        else:
            raise KeyError(name)

    def items(self):
        result = []
        for name in self._PROPERTIES.get(self.__class__, set()):
            try:
                result.append((name, getattr(self, '_%s' % name)))
            except AttributeError:
                pass
        return result

    def keys(self):
        result = set()
        for name in self._PROPERTIES.get(self.__class__, set()):
            if hasattr(self, '_%s' % name):
                result.add(name)
        return result

    ######################################################################
    # Get the rendered form of the style declaration
    ######################################################################
    def __str__(self):
        non_default = []
        for name in self._PROPERTIES.get(self.__class__, set()):
            try:
                non_default.append((
                    name.replace('_', '-'),
                    getattr(self, '_%s' % name)
                ))
            except AttributeError:
                pass

        return "; ".join(
            "%s: %s" % (name, value)
            for name, value in sorted(non_default)
        )

    @classmethod
    def validated_property(cls, name, choices, initial=None):
        "Define a simple validated property attribute."
        try:
            initial = choices.validate(initial)
        except ValueError:
            raise ValueError("Invalid initial value '{}' for property '{}'".format(initial, name))

        def getter(self):
            return getattr(self, '_%s' % name, initial)

        def setter(self, value):
            try:
                value = choices.validate(value)
            except ValueError:
                raise ValueError("Invalid value '{}' for property '{}'; Valid values are: {}".format(
                    value, name, choices
                ))

            if value != getattr(self, '_%s' % name, initial):
                setattr(self, '_%s' % name, value)
                self.apply(name, value)

        def deleter(self):
            try:
                value = getattr(self, '_%s' % name, initial)
                delattr(self, '_%s' % name)
                if value != initial:
                    self.apply(name, initial)
            except AttributeError:
                # Attribute doesn't exist
                pass

        cls._PROPERTIES.setdefault(cls, set()).add(name)
        cls._ALL_PROPERTIES.setdefault(cls, set()).add(name)
        setattr(cls, name, property(getter, setter, deleter))

    @classmethod
    def directional_property(cls, name):
        "Define a property attribute that proxies for top/right/bottom/left alternatives."
        def getter(self):
            return (
                getattr(self, name % '_top'),
                getattr(self, name % '_right'),
                getattr(self, name % '_bottom'),
                getattr(self, name % '_left'),
            )

        def setter(self, value):
            if isinstance(value, tuple):
                if len(value) == 4:
                    setattr(self, name % '_top', value[0])
                    setattr(self, name % '_right', value[1])
                    setattr(self, name % '_bottom', value[2])
                    setattr(self, name % '_left', value[3])
                elif len(value) == 3:
                    setattr(self, name % '_top', value[0])
                    setattr(self, name % '_right', value[1])
                    setattr(self, name % '_bottom', value[2])
                    setattr(self, name % '_left', value[1])
                elif len(value) == 2:
                    setattr(self, name % '_top', value[0])
                    setattr(self, name % '_right', value[1])
                    setattr(self, name % '_bottom', value[0])
                    setattr(self, name % '_left', value[1])
                elif len(value) == 1:
                    setattr(self, name % '_top', value[0])
                    setattr(self, name % '_right', value[0])
                    setattr(self, name % '_bottom', value[0])
                    setattr(self, name % '_left', value[0])
                else:
                    raise ValueError("Invalid value for '{}'; value must be an number, or a 1-4 tuple.".format(name % ''))
            else:
                setattr(self, name % '_top', value)
                setattr(self, name % '_right', value)
                setattr(self, name % '_bottom', value)
                setattr(self, name % '_left', value)

        def deleter(self):
            delattr(self, name % '_top')
            delattr(self, name % '_right')
            delattr(self, name % '_bottom')
            delattr(self, name % '_left')

        cls._ALL_PROPERTIES.setdefault(cls, set()).add(name % '')
        setattr(cls, name % '', property(getter, setter, deleter))

    # def list_property(name, choices, initial=None):
    #     "Define a property attribute that accepts a list of independently validated values."
    #     initial = choices.validate(initial)

    #     def getter(self):
    #         return getattr(self, '_%s' % name, initial)

    #     def setter(self, values):
    #         try:
    #             value = [choices.validate(v) for v in values.split(',')]
    #         except ValueError:
    #             raise ValueError("Invalid value in for list property '%s'; Valid values are: %s" % (
    #                 name, choices
    #             ))

    #         if value != getattr(self, '_%s' % name, initial):
    #             setattr(self, '_%s' % name, value)
    #             self.apply(name, value)

    #     def deleter(self):
    #         try:
    #             delattr(self, '_%s' % name)
    #             self.apply(name, value)
    #         except AttributeError:
    #             # Attribute doesn't exist
    #             pass

    #     _PROPERTIES.add(name)
    #     return property(getter, setter, deleter)




from unittest import TestCase

from travertino.declaration import BaseStyle
from travertino.layout import BaseBox, Viewport
from travertino.node import Node
from travertino.size import BaseIntrinsicSize


class Style(BaseStyle):
    class IntrinsicSize(BaseIntrinsicSize):
        pass

    class Box(BaseBox):
        pass

    def layout(self, root, viewport):
        # A simple layout scheme that allocats twice the viewport size.
        root.layout.content_width = viewport.width * 2
        root.layout.content_height = viewport.height * 2


class NodeTests(TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_create_leaf(self):
        "A leaf can be created"
        style = Style()
        leaf = Node(style=style)

        self.assertIsNone(leaf._children)
        self.assertEqual(leaf.children, [])
        self.assertFalse(leaf.can_have_children)

        # An unattached leaf is a root
        self.assertIsNone(leaf.parent)
        self.assertEqual(leaf.root, leaf)

        # A leaf can't have children
        child = Node(style=style)

        with self.assertRaises(ValueError):
            leaf.add(child)

    def test_create_node(self):
        "A node can be created with children"
        style = Style()

        child1 = Node(style=style)
        child2 = Node(style=style)
        child3 = Node(style=style)

        node = Node(style=style, children=[child1, child2, child3])

        self.assertEqual(node.children, [child1, child2, child3])
        self.assertTrue(node.can_have_children)

        # The node is the root as well.
        self.assertIsNone(node.parent)
        self.assertEqual(node.root, node)

        # The children all point at the node.
        self.assertEqual(child1.parent, node)
        self.assertEqual(child1.root, node)

        self.assertEqual(child2.parent, node)
        self.assertEqual(child2.root, node)

        self.assertEqual(child3.parent, node)
        self.assertEqual(child3.root, node)

        # Create another node
        new_node = Node(style=style, children=[])

        self.assertEqual(new_node.children, [])
        self.assertTrue(new_node.can_have_children)

        # Add the old node as a child of the new one.
        new_node.add(node)

        # The new node is the root
        self.assertIsNone(new_node.parent)
        self.assertEqual(new_node.root, new_node)

        # The node is the root as well.
        self.assertEqual(node.parent, new_node)
        self.assertEqual(node.root, new_node)

        # The children all point at the node.
        self.assertEqual(child1.parent, node)
        self.assertEqual(child1.root, new_node)

        self.assertEqual(child2.parent, node)
        self.assertEqual(child2.root, new_node)

        self.assertEqual(child3.parent, node)
        self.assertEqual(child3.root, new_node)

    def test_refresh(self):
        "The layout can be refreshed, and the applicator invoked"
        # Define an applicator that tracks the node being rendered and it's size
        class Applicator:
            def __init__(self, node):
                self.tasks = []
                self.node = node

            def set_bounds(self):
                self.tasks.append((self.node, self.node.layout.content_width, self.node.layout.content_height))

        class TestNode(Node):
            def __init__(self, style, children=None):
                super().__init__(
                    style=style,
                    applicator=Applicator(self),
                    children=children
                )

        # Define a simple 2 level tree of nodes.
        style = Style()
        child1 = TestNode(style=style)
        child2 = TestNode(style=style)
        child3 = TestNode(style=style)

        node = TestNode(style=style, children=[child1, child2, child3])

        # Refresh the root node
        node.refresh(Viewport(width=10, height=20))

        # Check the output is as expected
        self.assertEqual(node.applicator.tasks, [(node, 20, 40)])
        self.assertEqual(child1.applicator.tasks, [])
        self.assertEqual(child2.applicator.tasks, [])
        self.assertEqual(child3.applicator.tasks, [])

        # Reset the applicator
        node.applicator.tasks = []

        # Refresh a child node
        child1.refresh(Viewport(width=15, height=25))

        # The root node was rendered, not the child.
        self.assertEqual(node.applicator.tasks, [(node, 30, 50)])
        self.assertEqual(child1.applicator.tasks, [])
        self.assertEqual(child2.applicator.tasks, [])
        self.assertEqual(child3.applicator.tasks, [])



from unittest import TestCase
from unittest.mock import Mock

from travertino.size import BaseIntrinsicSize, at_least


class SizeTests(TestCase):
    def assertSize(self, size, values):
        self.assertEqual(values[0], size.width)
        self.assertEqual(values[1], size.height)
        self.assertEqual(values[2], size.ratio)

    def setUp(self):
        self.maxDiff = None

        self.layout = Mock()
        self.size = BaseIntrinsicSize(layout=self.layout)
        self.size._width = 1
        self.size._height = 2
        self.size._ratio = 0.1

        self.assertSize(self.size, (1, 2, 0.1))

    def test_at_least_repr(self):
        self.assertEqual(repr(at_least(10)), 'at least 10')

    def test_size_repr(self):
        self.assertEqual(repr(self.size), '(1, 2)')
        self.size.width = at_least(10)
        self.assertEqual(repr(self.size), '(at least 10, 2)')

    def test_set_width(self):
        self.size.width = 10
        self.assertSize(self.size, (10, 2, 0.1))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_width=10)

        # Clean the layout
        self.layout.dirty.reset_mock()

        # Set the width to the same value
        self.size.width = 10
        self.assertSize(self.size, (10, 2, 0.1))

        # Layout has NOT been dirtied.
        self.layout.dirty.assert_not_called()

        # Set the width to something new
        self.size.width = 20
        self.assertSize(self.size, (20, 2, 0.1))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_width=20)

    def test_set_height(self):
        self.size.height = 10
        self.assertSize(self.size, (1, 10, 0.1))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_height=10)

        # Clean the layout
        self.layout.dirty.reset_mock()

        # Set the height to the same value
        self.size.height = 10
        self.assertSize(self.size, (1, 10, 0.1))

        # Layout has NOT been dirtied.
        self.layout.dirty.assert_not_called()

        # Set the height to something new
        self.size.height = 20
        self.assertSize(self.size, (1, 20, 0.1))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_height=20)

    def test_set_min_width(self):
        self.size.width = at_least(10)
        self.assertSize(self.size, (at_least(10), 2, 0.1))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_width=at_least(10))

        # Clean the layout
        self.layout.dirty.reset_mock()

        # Set the exact_width to the same value
        self.size.width = at_least(10)
        self.assertSize(self.size, (at_least(10), 2, 0.1))

        # Layout has NOT been dirtied.
        self.layout.dirty.assert_not_called()

        # Set the exact_width to the same value, but not as a minimum
        self.size.width = 10
        self.assertSize(self.size, (10, 2, 0.1))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_width=10)

        # Clean the layout
        self.layout.dirty.reset_mock()

        # Set the exact_width to something new
        self.size.width = at_least(20)
        self.assertSize(self.size, (at_least(20), 2, 0.1))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_width=at_least(20))

    def test_set_min_height(self):
        self.size.height = at_least(10)
        self.assertSize(self.size, (1, at_least(10), 0.1))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_height=at_least(10))

        # Clean the layout
        self.layout.dirty.reset_mock()

        # Set the exact height to the same value
        self.size.height = at_least(10)
        self.assertSize(self.size, (1, at_least(10), 0.1))

        # Layout has NOT been dirtied.
        self.layout.dirty.assert_not_called()

        # Set the exact_height to the same value, but not as a minimum
        self.size.height = 10
        self.assertSize(self.size, (1, 10, 0.1))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_height=10)

        # Clean the layout
        self.layout.dirty.reset_mock()

        # Set the exact height to something else
        self.size.height = at_least(20)
        self.assertSize(self.size, (1, at_least(20), 0.1))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_height=at_least(20))

    def test_set_ratio(self):
        self.size.ratio = 0.5
        self.assertSize(self.size, (1, 2, 0.5))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_ratio=0.5)

        # Clean the layout
        self.layout.dirty.reset_mock()

        # Set the ratio to the same value
        self.size.ratio = 0.5
        self.assertSize(self.size, (1, 2, 0.5))

        # Layout has NOT been dirtied.
        self.layout.dirty.assert_not_called()

        # Set the ratio to something else
        self.size.ratio = 0.75
        self.assertSize(self.size, (1, 2, 0.75))

        # Layout has been dirtied.
        self.layout.dirty.assert_called_once_with(intrinsic_ratio=0.75)



from unittest import TestCase
from unittest.mock import Mock

from travertino.colors import NAMED_COLOR, rgb
from travertino.constants import TOP, GOLDENROD, REBECCAPURPLE
from travertino.declaration import Choices, BaseStyle


class PropertyChoiceTests(TestCase):
    def assert_property(self, obj, value, check_mock=True):
        self.assertEqual(obj.prop, value)
        if check_mock:
            obj.apply.assert_called_once_with('prop', value)
            obj.apply.reset_mock()

    def test_none(self):
        class MyObject(BaseStyle):
            def __init__(self):
                self.apply = Mock()
        MyObject.validated_property('prop', choices=Choices(None), initial=None)

        obj = MyObject()
        self.assertIsNone(obj.prop)

        with self.assertRaises(ValueError):
            obj.prop = 10

        with self.assertRaises(ValueError):
            obj.prop = 3.14159

        with self.assertRaises(ValueError):
            obj.prop = REBECCAPURPLE

        with self.assertRaises(ValueError):
            obj.prop = '#112233'

        with self.assertRaises(ValueError):
            obj.prop = 'a'

        with self.assertRaises(ValueError):
            obj.prop = 'b'

        obj.prop = None
        self.assert_property(obj, None, check_mock=False)

        obj.prop = 'none'
        self.assert_property(obj, None, check_mock=False)

        # Check the error message
        try:
            obj.prop = 'invalid'
            self.fail('Should raise ValueError')
        except ValueError as v:
            self.assertEqual(
                str(v),
                "Invalid value 'invalid' for property 'prop'; Valid values are: none"
            )

    def test_allow_string(self):
        class MyObject(BaseStyle):
            def __init__(self):
                self.apply = Mock()
        MyObject.validated_property('prop', choices=Choices(string=True), initial='start')

        obj = MyObject()
        self.assertEqual(obj.prop, 'start')

        with self.assertRaises(ValueError):
            obj.prop = 10

        with self.assertRaises(ValueError):
            obj.prop = 3.14159

        obj.prop = REBECCAPURPLE
        self.assert_property(obj, 'rebeccapurple')

        obj.prop = '#112233'
        self.assert_property(obj, '#112233')

        obj.prop = 'a'
        self.assert_property(obj, 'a')

        obj.prop = 'b'
        self.assert_property(obj, 'b')

        with self.assertRaises(ValueError):
            obj.prop = None

        obj.prop = 'none'
        self.assert_property(obj, 'none')

        # Check the error message
        try:
            obj.prop = 99
            self.fail('Should raise ValueError')
        except ValueError as v:
            self.assertEqual(
                str(v),
                "Invalid value '99' for property 'prop'; Valid values are: <string>"
            )

    def test_allow_integer(self):
        class MyObject(BaseStyle):
            def __init__(self):
                self.apply = Mock()
        MyObject.validated_property('prop', choices=Choices(integer=True), initial=0)

        obj = MyObject()
        self.assertEqual(obj.prop, 0)

        obj.prop = 10
        self.assert_property(obj, 10)

        # This is an odd case; Python happily rounds floats to integers.
        # It's more trouble than it's worth to correct this.
        obj.prop = 3.14159
        self.assert_property(obj, 3)

        with self.assertRaises(ValueError):
            obj.prop = REBECCAPURPLE

        with self.assertRaises(ValueError):
            obj.prop = '#112233'

        with self.assertRaises(ValueError):
            obj.prop = 'a'

        with self.assertRaises(ValueError):
            obj.prop = 'b'

        with self.assertRaises(ValueError):
            obj.prop = None

        with self.assertRaises(ValueError):
            obj.prop = 'none'

        # Check the error message
        try:
            obj.prop = 'invalid'
            self.fail('Should raise ValueError')
        except ValueError as v:
            self.assertEqual(
                str(v),
                "Invalid value 'invalid' for property 'prop'; Valid values are: <integer>"
            )

    def test_allow_number(self):
        class MyObject(BaseStyle):
            def __init__(self):
                self.apply = Mock()
        MyObject.validated_property('prop', choices=Choices(number=True), initial=0)

        obj = MyObject()
        self.assertEqual(obj.prop, 0)

        obj.prop = 10
        self.assert_property(obj, 10.0)

        obj.prop = 3.14159
        self.assert_property(obj, 3.14159)

        with self.assertRaises(ValueError):
            obj.prop = REBECCAPURPLE

        with self.assertRaises(ValueError):
            obj.prop = '#112233'

        with self.assertRaises(ValueError):
            obj.prop = 'a'

        with self.assertRaises(ValueError):
            obj.prop = 'b'

        with self.assertRaises(ValueError):
            obj.prop = None

        with self.assertRaises(ValueError):
            obj.prop = 'none'

        # Check the error message
        try:
            obj.prop = 'invalid'
            self.fail('Should raise ValueError')
        except ValueError as v:
            self.assertEqual(
                str(v),
                "Invalid value 'invalid' for property 'prop'; Valid values are: <number>"
            )

    def test_allow_color(self):
        class MyObject(BaseStyle):
            def __init__(self):
                self.apply = Mock()
        MyObject.validated_property('prop', choices=Choices(color=True), initial='goldenrod')

        obj = MyObject()
        self.assertEqual(obj.prop, NAMED_COLOR[GOLDENROD])

        with self.assertRaises(ValueError):
            obj.prop = 10

        with self.assertRaises(ValueError):
            obj.prop = 3.14159

        obj.prop = REBECCAPURPLE
        self.assert_property(obj, NAMED_COLOR[REBECCAPURPLE])

        obj.prop = '#112233'
        self.assert_property(obj, rgb(0x11, 0x22, 0x33))

        with self.assertRaises(ValueError):
            obj.prop = 'a'

        with self.assertRaises(ValueError):
            obj.prop = 'b'

        with self.assertRaises(ValueError):
            obj.prop = None

        with self.assertRaises(ValueError):
            obj.prop = 'none'

        # Check the error message
        try:
            obj.prop = 'invalid'
            self.fail('Should raise ValueError')
        except ValueError as v:
            self.assertEqual(
                str(v),
                "Invalid value 'invalid' for property 'prop'; Valid values are: <color>"
            )

    def test_values(self):
        class MyObject(BaseStyle):
            def __init__(self):
                self.apply = Mock()
        MyObject.validated_property('prop', choices=Choices('a', 'b', None), initial='a')

        obj = MyObject()
        self.assertEqual(obj.prop, 'a')

        with self.assertRaises(ValueError):
            obj.prop = 10

        with self.assertRaises(ValueError):
            obj.prop = 3.14159

        with self.assertRaises(ValueError):
            obj.prop = REBECCAPURPLE

        with self.assertRaises(ValueError):
            obj.prop = '#112233'

        obj.prop = None
        self.assert_property(obj, None)

        obj.prop = 'a'
        self.assert_property(obj, 'a')

        obj.prop = 'none'
        self.assert_property(obj, None)

        obj.prop = 'b'
        self.assert_property(obj, 'b')


        # Check the error message
        try:
            obj.prop = 'invalid'
            self.fail('Should raise ValueError')
        except ValueError as v:
            self.assertEqual(
                str(v),
                "Invalid value 'invalid' for property 'prop'; Valid values are: a, b, none"
            )

    def test_multiple_choices(self):
        class MyObject(BaseStyle):
            def __init__(self):
                self.apply = Mock()
        MyObject.validated_property('prop', choices=Choices(
                'a', 'b', None,
                number=True, color=True
            ), initial=None)

        obj = MyObject()

        obj.prop = 10
        self.assert_property(obj, 10.0)

        obj.prop = 3.14159
        self.assert_property(obj, 3.14159)

        obj.prop = REBECCAPURPLE
        self.assert_property(obj, NAMED_COLOR[REBECCAPURPLE])

        obj.prop = '#112233'
        self.assert_property(obj, rgb(0x11, 0x22, 0x33))

        obj.prop = None
        self.assert_property(obj, None)

        obj.prop = 'a'
        self.assert_property(obj, 'a')

        obj.prop = 'none'
        self.assert_property(obj, None)

        obj.prop = 'b'
        self.assert_property(obj, 'b')

        # Check the error message
        try:
            obj.prop = 'invalid'
            self.fail('Should raise ValueError')
        except ValueError as v:
            self.assertEqual(
                str(v),
                "Invalid value 'invalid' for property 'prop'; "
                "Valid values are: a, b, none, <number>, <color>"
            )

    def test_string_symbol(self):
        class MyObject(BaseStyle):
            def __init__(self):
                self.apply = Mock()
        MyObject.validated_property('prop', choices=Choices(TOP, None), initial=None)

        obj = MyObject()

        # Set a symbolic value using the string value of the symbol
        # We can't just use the string directly, though - that would
        # get optimized by the compiler. So we create a string and
        # transform it into the value we want.
        val = 'TOP'
        obj.prop = val.lower()

        # Both equality and instance checking should work.
        self.assertEqual(obj.prop, TOP)
        self.assertIs(obj.prop, TOP)



from unittest import TestCase

from travertino.colors import color, hsl, hsla, rgb, rgba


class ColorTests(TestCase):
    def assertEqualColor(self, a, b):
        self.assertEqual(a.rgba.r, b.rgba.r)
        self.assertEqual(a.rgba.g, b.rgba.g)
        self.assertEqual(a.rgba.b, b.rgba.b)
        self.assertEqual(a.rgba.a, b.rgba.a)

    def test_rgb_repr(self):
        self.assertEqual(repr(rgb(10, 20, 30)), "rgb(10, 20, 30)")

    def test_rgba_repr(self):
        self.assertEqual(repr(rgba(10, 20, 30, 0.5)), "rgba(10, 20, 30, 0.5)")

    def test_hsl_repr(self):
        self.assertEqual(repr(hsl(10, 0.2, 0.3)), "hsl(10, 0.2, 0.3)")

    def test_hsla_repr(self):
        self.assertEqual(repr(hsla(10, 0.2, 0.3, 0.5)), "hsla(10, 0.2, 0.3, 0.5)")

    def test_hsl_blacks(self):
        self.assertEqualColor(hsl(0, 0.0, 0.0), rgb(0x00, 0x00, 0x00))
        self.assertEqualColor(hsl(60, 0.0, 0.0), rgb(0x00, 0x00, 0x00))
        self.assertEqualColor(hsl(180, 0.0, 0.0), rgb(0x00, 0x00, 0x00))
        self.assertEqualColor(hsl(240, 0.0, 0.0), rgb(0x00, 0x00, 0x00))
        self.assertEqualColor(hsl(360, 0.0, 0.0), rgb(0x00, 0x00, 0x00))

    def test_hsl_whites(self):
        self.assertEqualColor(hsl(0, 0.0, 1.0), rgb(0xff, 0xff, 0xff))
        self.assertEqualColor(hsl(60, 0.0, 1.0), rgb(0xff, 0xff, 0xff))
        self.assertEqualColor(hsl(180, 0.0, 1.0), rgb(0xff, 0xff, 0xff))
        self.assertEqualColor(hsl(240, 0.0, 1.0), rgb(0xff, 0xff, 0xff))
        self.assertEqualColor(hsl(360, 0.0, 1.0), rgb(0xff, 0xff, 0xff))

    def test_hsl_grays(self):
        self.assertEqualColor(hsl(0, 0.0, 0.2), rgb(0x33, 0x33, 0x33))
        self.assertEqualColor(hsl(0, 0.0, 0.4), rgb(0x66, 0x66, 0x66))
        self.assertEqualColor(hsl(0, 0.0, 0.5), rgb(0x80, 0x80, 0x80))
        self.assertEqualColor(hsl(0, 0.0, 0.6), rgb(0x99, 0x99, 0x99))
        self.assertEqualColor(hsl(0, 0.0, 0.8), rgb(0xcc, 0xcc, 0xcc))

    def test_hsl_primaries(self):
        self.assertEqualColor(hsl(0, 1.0, 0.5), rgb(0xff, 0x00, 0x00))
        self.assertEqualColor(hsl(60, 1.0, 0.5), rgb(0xff, 0xff, 0x00))
        self.assertEqualColor(hsl(120, 1.0, 0.5), rgb(0x00, 0xff, 0x00))
        self.assertEqualColor(hsl(180, 1.0, 0.5), rgb(0x00, 0xff, 0xff))
        self.assertEqualColor(hsl(240, 1.0, 0.5), rgb(0x00, 0x00, 0xff))
        self.assertEqualColor(hsl(300, 1.0, 0.5), rgb(0xff, 0x00, 0xff))
        self.assertEqualColor(hsl(360, 1.0, 0.5), rgb(0xff, 0x00, 0x00))

    def test_hsl_muted(self):
        self.assertEqualColor(hsl(0, 0.25, 0.25), rgb(0x50, 0x30, 0x30))
        self.assertEqualColor(hsl(60, 0.25, 0.25), rgb(0x50, 0x50, 0x30))
        self.assertEqualColor(hsl(120, 0.25, 0.25), rgb(0x30, 0x50, 0x30))
        self.assertEqualColor(hsl(180, 0.25, 0.25), rgb(0x30, 0x50, 0x50))
        self.assertEqualColor(hsl(240, 0.25, 0.25), rgb(0x30, 0x30, 0x50))
        self.assertEqualColor(hsl(300, 0.25, 0.25), rgb(0x50, 0x30, 0x50))
        self.assertEqualColor(hsl(360, 0.25, 0.25), rgb(0x50, 0x30, 0x30))

        self.assertEqualColor(hsl(0, 0.25, 0.75), rgb(0xcf, 0xaf, 0xaf))
        self.assertEqualColor(hsl(60, 0.25, 0.75), rgb(0xcf, 0xcf, 0xaf))
        self.assertEqualColor(hsl(120, 0.25, 0.75), rgb(0xaf, 0xcf, 0xaf))
        self.assertEqualColor(hsl(180, 0.25, 0.75), rgb(0xaf, 0xcf, 0xcf))
        self.assertEqualColor(hsl(240, 0.25, 0.75), rgb(0xaf, 0xaf, 0xcf))
        self.assertEqualColor(hsl(300, 0.25, 0.75), rgb(0xcf, 0xaf, 0xcf))
        self.assertEqualColor(hsl(360, 0.25, 0.75), rgb(0xcf, 0xaf, 0xaf))

        self.assertEqualColor(hsl(0, 0.75, 0.75), rgb(0xef, 0x8f, 0x8f))
        self.assertEqualColor(hsl(60, 0.75, 0.75), rgb(0xef, 0xef, 0x8f))
        self.assertEqualColor(hsl(120, 0.75, 0.75), rgb(0x8f, 0xef, 0x8f))
        self.assertEqualColor(hsl(180, 0.75, 0.75), rgb(0x8f, 0xef, 0xef))
        self.assertEqualColor(hsl(240, 0.75, 0.75), rgb(0x8f, 0x8f, 0xef))
        self.assertEqualColor(hsl(300, 0.75, 0.75), rgb(0xef, 0x8f, 0xef))
        self.assertEqualColor(hsl(360, 0.75, 0.75), rgb(0xef, 0x8f, 0x8f))

        self.assertEqualColor(hsl(0, 0.75, 0.25), rgb(0x70, 0x10, 0x10))
        self.assertEqualColor(hsl(60, 0.75, 0.25), rgb(0x70, 0x70, 0x10))
        self.assertEqualColor(hsl(120, 0.75, 0.25), rgb(0x10, 0x70, 0x10))
        self.assertEqualColor(hsl(180, 0.75, 0.25), rgb(0x10, 0x70, 0x70))
        self.assertEqualColor(hsl(240, 0.75, 0.25), rgb(0x10, 0x10, 0x70))
        self.assertEqualColor(hsl(300, 0.75, 0.25), rgb(0x70, 0x10, 0x70))
        self.assertEqualColor(hsl(360, 0.75, 0.25), rgb(0x70, 0x10, 0x10))

    def test_hsl_alpha(self):
        self.assertEqualColor(hsla(60, 0.0, 0.0, 0.3), rgba(0x00, 0x00, 0x00, 0.3))
        self.assertEqualColor(hsla(60, 0.0, 1.0, 0.3), rgba(0xff, 0xff, 0xff, 0.3))
        self.assertEqualColor(hsla(60, 1.0, 0.5, 0.3), rgba(0xff, 0xff, 0x00, 0.3))
        self.assertEqualColor(hsla(60, 0.25, 0.25, 0.3), rgba(0x50, 0x50, 0x30, 0.3))
        self.assertEqualColor(hsla(60, 0.25, 0.75, 0.3), rgba(0xcf, 0xcf, 0xaf, 0.3))
        self.assertEqualColor(hsla(60, 0.75, 0.75, 0.3), rgba(0xef, 0xef, 0x8f, 0.3))
        self.assertEqualColor(hsla(60, 0.75, 0.25, 0.3), rgba(0x70, 0x70, 0x10, 0.3))


class ParseColorTests(TestCase):
    def assertEqualHSL(self, value, expected):
        # Nothing fancy - a color is equal if the attributes are all the same
        actual = color(value)
        self.assertEqual(actual.h, expected.h)
        self.assertEqual(actual.s, expected.s)
        self.assertEqual(actual.l, expected.l)
        self.assertAlmostEqual(actual.a, expected.a, places=3)

    def assertEqualColor(self, value, expected):
        # Nothing fancy - a color is equal if the attributes are all the same
        actual = color(value)
        self.assertEqual(actual.r, expected.r)
        self.assertEqual(actual.g, expected.g)
        self.assertEqual(actual.b, expected.b)
        self.assertAlmostEqual(actual.a, expected.a, places=3)

    def test_noop(self):
        self.assertEqualColor(rgba(1, 2, 3, 0.5), rgba(1, 2, 3, 0.5))
        self.assertEqualHSL(hsl(1, 0.2, 0.3), hsl(1, 0.2, 0.3))

    def test_rgb(self):
        self.assertEqualColor('rgb(1,2,3)', rgb(1, 2, 3))
        self.assertEqualColor('rgb(1, 2, 3)', rgb(1, 2, 3))
        self.assertEqualColor('rgb( 1 , 2 , 3)', rgb(1, 2, 3))

        self.assertEqualColor('#123', rgb(0x11, 0x22, 0x33))
        self.assertEqualColor('#112233', rgb(0x11, 0x22, 0x33))
        self.assertEqualColor('#abc', rgb(0xaa, 0xbb, 0xcc))
        self.assertEqualColor('#ABC', rgb(0xaa, 0xbb, 0xcc))
        self.assertEqualColor('#abcdef', rgb(0xab, 0xcd, 0xef))
        self.assertEqualColor('#ABCDEF', rgb(0xab, 0xcd, 0xef))

        with self.assertRaises(ValueError):
            color('rgb(10, 20)')

        with self.assertRaises(ValueError):
            color('rgb(a, 10, 20)')

        with self.assertRaises(ValueError):
            color('rgb(10, b, 20)')

        with self.assertRaises(ValueError):
            color('rgb(10, 20, c)')

        with self.assertRaises(ValueError):
            color('rgb(10, 20, 30, 0.5)')

    def test_rgba(self):
        self.assertEqualColor('rgba(1,2,3,0.5)', rgba(1, 2, 3, 0.5))
        self.assertEqualColor('rgba(1, 2, 3, 0.5)', rgba(1, 2, 3, 0.5))
        self.assertEqualColor('rgba( 1 , 2 , 3 , 0.5)', rgba(1, 2, 3, 0.5))

        self.assertEqualColor('#1234', rgba(0x11, 0x22, 0x33, 0.2666))
        self.assertEqualColor('#11223344', rgba(0x11, 0x22, 0x33, 0.2666))
        self.assertEqualColor('#abcd', rgba(0xaa, 0xbb, 0xcc, 0.8666))
        self.assertEqualColor('#ABCD', rgba(0xaa, 0xbb, 0xcc, 0.8666))
        self.assertEqualColor('#abcdefba', rgba(0xab, 0xcd, 0xef, 0.7294))
        self.assertEqualColor('#ABCDEFBA', rgba(0xab, 0xcd, 0xef, 0.7294))

        with self.assertRaises(ValueError):
            color('rgba(10, 20, 30)')

        with self.assertRaises(ValueError):
            color('rgba(a, 10, 20, 0.5)')

        with self.assertRaises(ValueError):
            color('rgba(10, b, 20, 0.5)')

        with self.assertRaises(ValueError):
            color('rgba(10, 20, c, 0.5)')

        with self.assertRaises(ValueError):
            color('rgba(10, 20, 30, c)')

        with self.assertRaises(ValueError):
            color('rgba(10, 20, 30, 0.5, 5)')

    def test_hsl(self):
        self.assertEqualHSL('hsl(1,20%,30%)', hsl(1, 0.2, 0.3))
        self.assertEqualHSL('hsl(1, 20%, 30%)', hsl(1, 0.2, 0.3))
        self.assertEqualHSL('hsl( 1, 20% , 30%)', hsl(1, 0.2, 0.3))

        with self.assertRaises(ValueError):
            color('hsl(1, 20%)')

        with self.assertRaises(ValueError):
            color('hsl(a, 20%, 30%)')

        with self.assertRaises(ValueError):
            color('hsl(1, a, 30%)')

        with self.assertRaises(ValueError):
            color('hsl(1, 20%, a)')

        with self.assertRaises(ValueError):
            color('hsl(1, 20%, 30%, 0.5)')

    def test_hsla(self):
        self.assertEqualHSL('hsla(1,20%,30%,0.5)', hsla(1, 0.2, 0.3, 0.5))
        self.assertEqualHSL('hsla(1, 20%, 30%, 0.5)', hsla(1, 0.2, 0.3, 0.5))
        self.assertEqualHSL('hsla( 1, 20% , 30% , 0.5)', hsla(1, 0.2, 0.3, 0.5))

        with self.assertRaises(ValueError):
            color('hsla(1, 20%, 30%)')

        with self.assertRaises(ValueError):
            color('hsla(a, 20%, 30%, 0.5)')

        with self.assertRaises(ValueError):
            color('hsla(1, a, 30%, 0.5)')

        with self.assertRaises(ValueError):
            color('hsla(1, 20%, a, 0.5)')

        with self.assertRaises(ValueError):
            color('hsla(1, 20%, 30%, a)')

        with self.assertRaises(ValueError):
            color('hsla(1, 20%, 30%, 0.5, 5)')

    def test_named_color(self):
        self.assertEqualColor('Red', rgb(0xFF, 0, 0))
        self.assertEqualColor('RED', rgb(0xFF, 0, 0))
        self.assertEqualColor('red', rgb(0xFF, 0, 0))
        self.assertEqualColor('rEd', rgb(0xFF, 0, 0))

        self.assertEqualColor('CornflowerBlue', rgb(0x64, 0x95, 0xED))
        self.assertEqualColor('cornflowerblue', rgb(0x64, 0x95, 0xED))
        self.assertEqualColor('CORNFLOWERBLUE', rgb(0x64, 0x95, 0xED))
        self.assertEqualColor('Cornflowerblue', rgb(0x64, 0x95, 0xED))
        self.assertEqualColor('CoRnFlOwErBlUe', rgb(0x64, 0x95, 0xED))

        with self.assertRaises(ValueError):
            color('not a color')



from unittest import TestCase
from unittest.mock import Mock, call

from travertino.declaration import BaseStyle, Choices


VALUE1 = 'value1'
VALUE2 = 'value2'
VALUE3 = 'value3'
VALUE_CHOICES = Choices(VALUE1, VALUE2, VALUE3, None, integer=True)
DEFAULT_VALUE_CHOICES = Choices(VALUE1, VALUE2, VALUE3, integer=True, default=True)


class Style(BaseStyle):
    def __init__(self, **kwargs):
        self.apply = Mock()
        super().__init__(**kwargs)


# Some properties with explicit initial values
Style.validated_property('explicit_const', choices=VALUE_CHOICES, initial=VALUE1)
Style.validated_property('explicit_value', choices=VALUE_CHOICES, initial=0)
Style.validated_property('explicit_none', choices=VALUE_CHOICES, initial=None)

# A property with an implicit default value.
# This usually means the default is platform specific.
Style.validated_property('implicit', choices=DEFAULT_VALUE_CHOICES)

# A set of directional properties
Style.validated_property('thing_top', choices=VALUE_CHOICES, initial=0)
Style.validated_property('thing_right', choices=VALUE_CHOICES, initial=0)
Style.validated_property('thing_bottom', choices=VALUE_CHOICES, initial=0)
Style.validated_property('thing_left', choices=VALUE_CHOICES, initial=0)
Style.directional_property('thing%s')


class TestNode:
    def __init__(self, style=None):
        if style is None:
            self.style = Style()
        else:
            self.style = style.copy(self)


class DeclarationTests(TestCase):
    def test_invalid_style(self):
        with self.assertRaises(ValueError):
            # Define a style that has an invalid initial value on a validated property
            class BadStyle(BaseStyle):
                pass
            BadStyle.validated_property('value', choices=VALUE_CHOICES, initial='something')

    def test_create_and_copy(self):
        style = Style(explicit_const=VALUE2, implicit=VALUE3)

        dup = style.copy()
        self.assertEqual(dup.explicit_const, VALUE2)
        self.assertEqual(dup.explicit_value, 0)
        self.assertEqual(dup.implicit, VALUE3)

    def test_reapply(self):
        node = TestNode(style=Style(explicit_const=VALUE2, implicit=VALUE3))

        node.style.reapply()
        node.style.apply.assert_has_calls([
            call('explicit_const', VALUE2),
            call('explicit_value', 0),
            call('explicit_none', None),
            call('implicit', VALUE3),
            call('thing_left', 0),
            call('thing_top', 0),
            call('thing_right', 0),
            call('thing_bottom', 0),
        ], any_order=True)


    def test_property_with_explicit_const(self):
        node = TestNode()

        # Default value is VALUE1
        self.assertIs(node.style.explicit_const, VALUE1)
        node.style.apply.assert_not_called()

        # Modify the value
        node.style.explicit_const = 10

        self.assertEqual(node.style.explicit_const, 10)
        node.style.apply.assert_called_once_with('explicit_const', 10)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Set the value to the same value.
        # No dirty notification is sent
        node.style.explicit_const = 10
        self.assertEqual(node.style.explicit_const, 10)
        node.style.apply.assert_not_called()

        # Set the value to something new
        # A dirty notification is set.
        node.style.explicit_const = 20
        self.assertEqual(node.style.explicit_const, 20)
        node.style.apply.assert_called_once_with('explicit_const', 20)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Clear the property
        del node.style.explicit_const
        self.assertIs(node.style.explicit_const, VALUE1)
        node.style.apply.assert_called_once_with('explicit_const', VALUE1)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Clear the property again.
        # The underlying attribute won't exist, so this
        # should be a no-op.
        del node.style.explicit_const
        self.assertIs(node.style.explicit_const, VALUE1)
        node.style.apply.assert_not_called()

    def test_property_with_explicit_value(self):
        node = TestNode()

        # Default value is 0
        self.assertEqual(node.style.explicit_value, 0)
        node.style.apply.assert_not_called()

        # Modify the value
        node.style.explicit_value = 10

        self.assertEqual(node.style.explicit_value, 10)
        node.style.apply.assert_called_once_with('explicit_value', 10)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Set the value to the same value.
        # No dirty notification is sent
        node.style.explicit_value = 10
        self.assertEqual(node.style.explicit_value, 10)
        node.style.apply.assert_not_called()

        # Set the value to something new
        # A dirty notification is set.
        node.style.explicit_value = 20
        self.assertEqual(node.style.explicit_value, 20)
        node.style.apply.assert_called_once_with('explicit_value', 20)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Clear the property
        del node.style.explicit_value
        self.assertEqual(node.style.explicit_value, 0)
        node.style.apply.assert_called_once_with('explicit_value', 0)

    def test_property_with_explicit_none(self):
        node = TestNode()

        # Default value is None
        self.assertIsNone(node.style.explicit_none)
        node.style.apply.assert_not_called()

        # Modify the value
        node.style.explicit_none = 10

        self.assertEqual(node.style.explicit_none, 10)
        node.style.apply.assert_called_once_with('explicit_none', 10)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Set the property to the same value.
        # No dirty notification is sent
        node.style.explicit_none = 10
        self.assertEqual(node.style.explicit_none, 10)
        node.style.apply.assert_not_called()

        # Set the property to something new
        # A dirty notification is set.
        node.style.explicit_none = 20
        self.assertEqual(node.style.explicit_none, 20)
        node.style.apply.assert_called_once_with('explicit_none', 20)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Clear the property
        del node.style.explicit_none
        self.assertIsNone(node.style.explicit_none)
        node.style.apply.assert_called_once_with('explicit_none', None)

    def test_property_with_implicit_default(self):
        node = TestNode()

        # Default value is None
        self.assertIsNone(node.style.implicit)
        node.style.apply.assert_not_called()

        # Modify the value
        node.style.implicit = 10

        self.assertEqual(node.style.implicit, 10)
        node.style.apply.assert_called_once_with('implicit', 10)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Set the value to the same value.
        # No dirty notification is sent
        node.style.implicit = 10
        self.assertEqual(node.style.implicit, 10)
        node.style.apply.assert_not_called()

        # Set the value to something new
        # A dirty notification is set.
        node.style.implicit = 20
        self.assertEqual(node.style.implicit, 20)
        node.style.apply.assert_called_once_with('implicit', 20)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Clear the property
        del node.style.implicit
        self.assertIsNone(node.style.implicit)
        node.style.apply.assert_called_once_with('implicit', None)

    def test_directional_property(self):
        node = TestNode()

        # Default value is 0
        self.assertEqual(node.style.thing, (0, 0, 0, 0))
        self.assertEqual(node.style.thing_top, 0)
        self.assertEqual(node.style.thing_right, 0)
        self.assertEqual(node.style.thing_bottom, 0)
        self.assertEqual(node.style.thing_left, 0)
        node.style.apply.assert_not_called()

        # Set a value in one axis
        node.style.thing_top = 10

        self.assertEqual(node.style.thing, (10, 0, 0, 0))
        self.assertEqual(node.style.thing_top, 10)
        self.assertEqual(node.style.thing_right, 0)
        self.assertEqual(node.style.thing_bottom, 0)
        self.assertEqual(node.style.thing_left, 0)
        node.style.apply.assert_called_once_with('thing_top', 10)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Set a value directly with a single item
        node.style.thing = (10,)

        self.assertEqual(node.style.thing, (10, 10, 10, 10))
        self.assertEqual(node.style.thing_top, 10)
        self.assertEqual(node.style.thing_right, 10)
        self.assertEqual(node.style.thing_bottom, 10)
        self.assertEqual(node.style.thing_left, 10)
        node.style.apply.assert_has_calls([
            call('thing_right', 10),
            call('thing_bottom', 10),
            call('thing_left', 10),
        ])

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Set a value directly with a single item
        node.style.thing = 30

        self.assertEqual(node.style.thing, (30, 30, 30, 30))
        self.assertEqual(node.style.thing_top, 30)
        self.assertEqual(node.style.thing_right, 30)
        self.assertEqual(node.style.thing_bottom, 30)
        self.assertEqual(node.style.thing_left, 30)
        node.style.apply.assert_has_calls([
            call('thing_top', 30),
            call('thing_right', 30),
            call('thing_bottom', 30),
            call('thing_left', 30),
        ])

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Set a value directly with a 2 values
        node.style.thing = (10, 20)

        self.assertEqual(node.style.thing, (10, 20, 10, 20))
        self.assertEqual(node.style.thing_top, 10)
        self.assertEqual(node.style.thing_right, 20)
        self.assertEqual(node.style.thing_bottom, 10)
        self.assertEqual(node.style.thing_left, 20)
        node.style.apply.assert_has_calls([
            call('thing_top', 10),
            call('thing_right', 20),
            call('thing_bottom', 10),
            call('thing_left', 20),
        ])

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Set a value directly with a 3 values
        node.style.thing = (10, 20, 30)

        self.assertEqual(node.style.thing, (10, 20, 30, 20))
        self.assertEqual(node.style.thing_top, 10)
        self.assertEqual(node.style.thing_right, 20)
        self.assertEqual(node.style.thing_bottom, 30)
        self.assertEqual(node.style.thing_left, 20)
        node.style.apply.assert_called_once_with('thing_bottom', 30)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Set a value directly with a 4 values
        node.style.thing = (10, 20, 30, 40)

        self.assertEqual(node.style.thing, (10, 20, 30, 40))
        self.assertEqual(node.style.thing_top, 10)
        self.assertEqual(node.style.thing_right, 20)
        self.assertEqual(node.style.thing_bottom, 30)
        self.assertEqual(node.style.thing_left, 40)
        node.style.apply.assert_called_once_with('thing_left', 40)

        # Set a value directly with an invalid number of values
        with self.assertRaises(ValueError):
            node.style.thing = ()

        with self.assertRaises(ValueError):
            node.style.thing = (10, 20, 30, 40, 50)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Clear a value on one axis
        del node.style.thing_top

        self.assertEqual(node.style.thing, (0, 20, 30, 40))
        self.assertEqual(node.style.thing_top, 0)
        self.assertEqual(node.style.thing_right, 20)
        self.assertEqual(node.style.thing_bottom, 30)
        self.assertEqual(node.style.thing_left, 40)
        node.style.apply.assert_called_once_with('thing_top', 0)

        # Restore the top thing
        node.style.thing_top = 10

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Clear a value directly
        del node.style.thing

        self.assertEqual(node.style.thing, (0, 0, 0, 0))
        self.assertEqual(node.style.thing_top, 0)
        self.assertEqual(node.style.thing_right, 0)
        self.assertEqual(node.style.thing_bottom, 0)
        self.assertEqual(node.style.thing_left, 0)
        node.style.apply.assert_has_calls([
            call('thing_right', 0),
            call('thing_bottom', 0),
            call('thing_left', 0),
        ])

    def test_set_multiple_properties(self):
        node = TestNode()

        # Set a pair of properties
        node.style.update(explicit_value=20, explicit_none=10)

        self.assertIs(node.style.explicit_const, VALUE1)
        self.assertEqual(node.style.explicit_none, 10)
        self.assertEqual(node.style.explicit_value, 20)
        node.style.apply.assert_has_calls([
                call('explicit_value', 20),
                call('explicit_none', 10),
            ], any_order=True)

        # Set a different pair of properties
        node.style.update(explicit_const=VALUE2, explicit_value=30)

        self.assertIs(node.style.explicit_const, VALUE2)
        self.assertEqual(node.style.explicit_value, 30)
        self.assertEqual(node.style.explicit_none, 10)
        node.style.apply.assert_has_calls([
                call('explicit_const', VALUE2),
                call('explicit_value', 30),
            ], any_order=True)

        # Clear the applicator mock
        node.style.apply.reset_mock()

        # Setting a non-property
        with self.assertRaises(NameError):
            node.style.update(not_a_property=10)

        node.style.apply.assert_not_called()

    def test_str(self):
        node = TestNode()

        node.style.update(
            explicit_const=VALUE2,
            explicit_value=20,
            thing=(30, 40, 50, 60),
        )

        self.assertEqual(
            str(node.style),
            "explicit-const: value2; "
            "explicit-value: 20; "
            "thing-bottom: 50; "
            "thing-left: 60; "
            "thing-right: 40; "
            "thing-top: 30"
        )

    def test_dict(self):
        "Style declarations expose a dict-like interface"
        node = TestNode()

        node.style.update(
            explicit_const=VALUE2,
            explicit_value=20,
            thing=(30, 40, 50, 60),
        )

        self.assertEqual(
            node.style.keys(),
            {'explicit_const', 'explicit_value', 'thing_bottom', 'thing_left', 'thing_right', 'thing_top'}
        )
        self.assertEqual(
            sorted(node.style.items()),
            sorted([
                ('explicit_const', 'value2'),
                ('explicit_value', 20),
                ('thing_bottom', 50),
                ('thing_left', 60),
                ('thing_right', 40),
                ('thing_top', 30),
            ])
        )

        # A property can be set, retrieved and cleared using the attribute name
        node.style['thing-bottom'] = 10
        self.assertEqual(node.style['thing-bottom'], 10)
        del node.style['thing-bottom']
        self.assertEqual(node.style['thing-bottom'], 0)

        # A property can be set, retrieved and cleared using the Python attribute name
        node.style['thing_bottom'] = 10
        self.assertEqual(node.style['thing_bottom'], 10)
        del node.style['thing_bottom']
        self.assertEqual(node.style['thing_bottom'], 0)

        # Clearing a valid property isn't an error
        del node.style['thing_bottom']
        self.assertEqual(node.style['thing_bottom'], 0)

        # Non-existent properties raise KeyError
        with self.assertRaises(KeyError):
            node.style['no-such-property'] = 'no-such-value'

        with self.assertRaises(KeyError):
            node.style['no-such-property']

        with self.assertRaises(KeyError):
            del node.style['no-such-property']



#!/usr/bin/env python
import io
import re

from setuptools import find_packages, setup

with io.open('./travertino/__init__.py', encoding='utf8') as version_file:
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file.read(), re.M)
    if version_match:
        version = version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")


with io.open('README.rst', encoding='utf8') as readme:
    long_description = readme.read()


setup(
    name='travertino',
    version=version,
    description='Travertino is a set of constants and utilities for describing user interfaces',
    long_description=long_description,
    author='Russell Keith-Magee',
    author_email='russell@keith-magee.com',
    url='https://github.com/pybee/travertino',
    packages=find_packages(exclude=['tests', 'utils']),
    install_requires=[],
    license='New BSD',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development',
        'Topic :: Software Development :: User Interfaces',
    ],
    test_suite='tests'
)
